from . import Entity as ent
from . import Field as fld
from . import Options as opt

class Mesh:
    def __init__(self):
        self.points = {}
        self.points_count = 0
        self.curves = {}
        self.curves_count = 0
        self.curveloops = {}
        self.curveloops_count = 0
        self.surfaces = {}
        self.surfaces_count = 0
        self.surfaceloops = {}
        self.surfaceloops_count = 0
        self.volumes = {}
        self.volumes_count = 0
        self.regions = {}
        self.regions_count = 0
        self.fields = {}
        self.fields_count = 0
        self.groups = {}
        self.groups_count = 0
        self.Options = opt.OptionsHolder()
        self.BackgroundField = None
        self.BoundaryLayerField = None
        self.Coherence = False

    def getPointsFromIndex(self, index):
        if isinstance(index, int):
            index = [index]
        points = []
        for i in index:
            points += [self.points[i]]
        return points

    def getCurvesFromIndex(self, index):
        if isinstance(index, int):
            index = [index]
        curves = []
        for i in index:
            curves += [self.curves[i]]
        return curves

    def getSurfacesFromIndex(self, index):
        if isinstance(index, int):
            index = [index]
        surfaces = []
        for i in index:
            surfaces += [self.surfaces[i]]
        return surfaces

    def getSurfaceLoopsFromIndex(self, index):
        if isinstance(index, int):
            index = [index]
        surfaceloops = []
        for i in index:
            surfaceloops += [self.surfaceloops[i]]
        return surfaceloops

    def getVolumesFromIndex(self, index):
        if isinstance(index, int):
            index = [index]
        volumes = []
        for i in index:
            volumes += [self.volumes[i]]
        return volumes

    def getFieldsFromIndex(self, index):
        if isinstance(index, int):
            index = [index]
        fields = []
        for i in index:
            fields += [self.fields[i]]
        return fields

    def getGroupsFromIndex(self, index):
        if isinstance(index, int):
            index = [index]
        groups = []
        for i in index:
            groups += [self.groups[i]]
        return groups

    def addEntity(self, entity):
        if isinstance(entity, ent.Point):
            if entity.nb is None:
                self.points_count += 1
                entity.nb = self.points_count
            assert not self.points.get(entity.nb), 'Point nb '+str(entity.nb)+' already exists!'
            self.points[entity.nb] = entity
        elif isinstance(entity, ent.CurveEntity):
            if entity.nb is None:
                self.curves_count += 1
                entity.nb = self.curves_count
            assert not self.curves.get(entity.nb), 'Curve nb '+str(entity.nb)+' already exists!'
            self.curves[entity.nb] = entity
        elif isinstance(entity, ent.CurveLoop):
            if entity.nb is None:
                self.curveloops_count += 1
                entity.nb = self.curveloops_count
            assert not self.curveloops.get(entity.nb), 'CurveLoop nb '+str(entity.nb)+' already exists!'
            self.curveloops[entity.nb] = entity
        elif isinstance(entity, ent.SurfaceEntity):
            if entity.nb is None:
                self.surfaces_count += 1
                entity.nb = self.surfaces_count
            assert not self.surfaces.get(entity.nb), 'Surface nb '+str(entity.nb)+' already exists!'
            self.surfaces[entity.nb] = entity
        elif isinstance(entity, ent.SurfaceLoop):
            if entity.nb is None:
                self.surfaceloops_count += 1
                entity.nb = self.surfaceloops_count
            assert not self.surfaceloops.get(entity.nb), 'SurfaceLoop nb '+str(entity.nb)+' already exists!'
            self.surfaceloops[entity.nb] = entity
        elif isinstance(entity, ent.VolumeEntity):
            if entity.nb is None:
                self.volumes_count += 1
                entity.nb = self.volumes_count
            assert not self.volumes.get(entity.nb), 'Volume nb '+str(entity.nb)+' already exists!'
            self.volumes[entity.nb] = entity
        elif isinstance(entity, ent.PhysicalGroup):
            self.addGroup(entity)
        elif isinstance(entity, fld.Field):
            self.addField(entity)
        else:
            raise TypeError("not a valid Entity instance")

    def addEntities(self, entities):
        for entity in entities:
            self.addEntity(entity)

    def addGroup(self, group):
        assert isinstance(group, ent.PhysicalGroup), 'Not a valid PhysicalGroup instance'
        if group.nb is None:
            self.groups_count += 1
            group.nb = self.groups_count
        assert not self.groups.get(group.nb), 'PhysicalGroup nb '+str(group.nb)+' already exists!'
        self.groups[group.nb] = group

    def addField(self, field):
        assert isinstance(field, fld.Field), 'Not a valid Field instance'
        if field.nb is None:
            self.fields_count += 1
            field.nb = self.fields_count
        assert not self.fields.get(field.nb), 'Field nb '+str(field.nb)+' already exists!'
        self.fields[field.nb] = field

    def setBackgroundField(self, field):
        if not self.fields.get(field.nb):
            self.fields[field.nb] = field
        self.BackgroundField = field

    def writeGeo(self, filename):
        geo = open(filename,'w')
        for key, entity in self.points.items():
            geo.write("{0}({1}) = {2};\n".format(entity.name, key, entity._val2str()))
        for key, entity in self.curves.items():
            geo.write("{0}({1}) = {2};\n".format(entity.name, key, entity._val2str()))
        for key, entity in self.curveloops.items():
            geo.write("{0}({1}) = {2};\n".format(entity.name, key, entity._val2str()))
        for key, entity in self.surfaces.items():
            geo.write("{0}({1}) = {2};\n".format(entity.name, key, entity._val2str()))
        for key, entity in self.surfaceloops.items():
            geo.write("{0}({1}) = {2};\n".format(entity.name, key, entity._val2str()))
        for key, entity in self.volumes.items():
            geo.write("{0}({1}) = {2};\n".format(entity.name, key, entity._val2str()))

        # Physical Groups
        geo.write('\n// Physical Groups\n')
        for i, group in self.groups.items():
            if group.name:
                name = '"'+group.name+'", '+str(group.nb)
            else:
                name = group.nb
            if group.points:
                points = []
                for key, point in group.points.items():
                    points.append(point.nb)
                geo.write("Physical Point({0}) = {{{1}}};\n".format(name, str(points)[1:-1]))
            if group.curves:
                curves = []
                for key, curve in group.curves.items():
                    curves.append(curve.nb)
                geo.write("Physical Curve({0}) = {{{1}}};\n".format(name, str(curves)[1:-1]))
            if group.surfaces:
                surfaces = []
                for key, surface in group.surfaces.items():
                    surfaces.append(surface.nb)
                geo.write("Physical Surface({0}) = {{{1}}};\n".format(name, str(surfaces)[1:-1]))
            if group.volumes:
                volumes = []
                for key, volume in group.volumes.items():
                    volumes.append(volume.nb)
                geo.write("Physical Volume({0}) = {{{1}}};\n".format(name, str(volumes)[1:-1]))

        for i, field in self.fields.items():
            geo.write('Field[{0}] = {1};\n'.format(field.nb, field.name))
            for attr in field.__dict__:
                val = getattr(field, attr)
                if val is not None:
                    if isinstance(val, str):
                        val_str = '"'+val+'"'
                    elif attr == 'EdgesList' or attr =='NodesList' or attr == 'FacesList' or attr == 'RegionsList' or attr == 'FieldsList' or attr == 'VerticesList':
                        val_str = '{'+str([v.nb for v in val])[1:-1]+'}'
                    elif attr == 'IField' or attr == 'FieldX' or attr == 'FieldY' or attr == 'FieldZ':
                        val_str = str(val.nb)
                    else:
                        val_str = str(val)
                        if isinstance(val, (list, tuple)):
                            val_str = '{'+val_str[1:-1]+'}' # replace () by {} in string
                    geo.write('Field[{0}].{1} = {2};\n'.format(field.nb, attr, val_str))

        if self.BackgroundField:
            geo.write("Background Field = {0};\n".format(self.BackgroundField.nb))
        if self.BoundaryLayerField:
            geo.write("BoundaryLayer Field = {0};\n".format(self.BoundaryLayerField.nb))

        def write_option(class_instance):
            class_name = class_instance.__class__.__name__
            for key in class_instance.__dict__:
                val = getattr(class_instance, key)
                if key != 'Color':
                    if val:
                        geo.write(class_name+'.'+key+'= {0};\n'.format(val))
                else:
                    for key2 in class_instance.Color.__dict__:
                        val2 = getattr(class_instance.Color, key2)
                        if val2:
                            geo.write(class_name+'.Color.'+key2+'= {0};\n'.format(val2))

        for key in self.Options.__dict__:
            options = getattr(self.Options, key)
            write_option(options)

        if self.Coherence:
            geo.write("Coherence;\n") # remove duplicates

        geo.close()


def geometry2mesh(domain):
    curves_dict = {}

    mesh = Mesh()

    if domain.boundaryTags:
        for tag, flag in domain.boundaryTags.items():
            phys = ent.PhysicalGroup(nb=flag, name=tag)
            mesh.addGroup(phys)

    for i, v in enumerate(domain.vertices):
        if domain.nd == 2:
            p = ent.Point([v[0], v[1], 0.])
        else:
            p = ent.Point(v)
        mesh.addEntity(p)
        g = mesh.groups.get(domain.vertexFlags[i])
        if g:
            g.addEntity(p)
    nb_points = i+1
    for i in range(nb_points):
        curves_dict[i] = {}

    for i, s in enumerate(domain.segments):
        curves_dict[s[0]][s[1]] = i
        l = ent.Curve([mesh.points[s[0]+1], mesh.points[s[1]+1]])
        mesh.addEntity(l)
        g = mesh.groups.get(domain.segmentFlags[i])
        if g:
            g.addEntity(l)

    for i, f in enumerate(domain.facets):
        if domain.nd == 3 or (domain.nd == 2 and i not in domain.holes_ind):
            curveloops = []
            for j, subf in enumerate(f):
                curveloop = []
                # vertices in facet
                for k, ver in enumerate(subf):
                    if ver in curves_dict[subf[k-1]].keys():
                        curveloop += [curves_dict[subf[k-1]][ver]+1]
                    elif subf[k-1] in curves_dict[ver].keys():
                        # reversed
                        curveloop += [(curves_dict[ver][subf[k-1]]+1)]
                    else:
                        l = ent.Curve([mesh.points[subf[k-1]+1], mesh.points[ver+1]])
                        mesh.addEntity(l)
                        curves_dict[ver][subf[k-1]] = l.nb-1
                        curves_dict[subf[k-1]][ver] = l.nb-1
                        curveloop += [l.nb]
                ll = ent.CurveLoop(mesh.getCurvesFromIndex(curveloop))
                mesh.addEntity(ll)
                curveloops += [ll.nb]
            s = ent.PlaneSurface([mesh.curveloops[loop] for loop in curveloops])
            mesh.addEntity(s)
            g = mesh.groups.get(domain.facetFlags[i])
            if g:
                g.addEntity(s)

    for i, V in enumerate(domain.volumes):
        surface_loops = []
        hole_loops = []
        for j, sV in enumerate(V):
            sl = ent.SurfaceLoop(mesh.getSurfacesFromIndex([sVnb + 1 for sVnb in sV]))
            mesh.addEntity(sl)
            surface_loops += [sl]
        vol = ent.Volume(surface_loops)
        mesh.addEntity(vol)
        g = mesh.groups.get(domain.regionFlags[i])
        if g:
            g.addEntity(vol)

    print("CREATED")
    return mesh



class PhysicalGroup(object):
    """PhysicalGroup

    Parameters
    ----------
    nb: Optional[int]
        Physical group number. If not set, it will use the count 
        number (incrementing automatically upon creation of new
        physical groups).
    name: Optional[str]
        Name of physical group.
    mesh: Optional[py2gmsh.Mesh.Mesh]
        Mesh class instance to which the group belongs.
    """
    def __init__(self, nb=None, name=None, mesh=None):
        self.nb = nb
        self.name = name
        self.points = {}
        self.curves = {}
        self.curveloops = {}
        self.surfaces = {}
        self.surfaceloops = {}
        self.volumes = {}
        self.regions = {}
        if mesh is not None:
            mesh.addGroup(self)

    def addEntity(self, entity):
        """Adds entity to PhysicalGroup instance

        Parameters
        ----------
        entity: py2gmsh.Entity.Entity
            Entity to add to group (e.g. Point, Curve, CurveLoop).
        """
        if isinstance(entity, Point):
            assert not self.points.get(entity.nb), 'Point nb '+str(entity.nb)+' already exists!'
            self.points[entity.nb] = entity
        elif isinstance(entity, CurveEntity):
            assert not self.curves.get(entity.nb), 'Curve nb '+str(entity.nb)+' already exists!'
            self.curves[entity.nb] = entity
        elif isinstance(entity, SurfaceEntity):
            assert not self.surfaces.get(entity.nb), 'Surface nb '+str(entity.nb)+' already exists!'
            self.surfaces[entity.nb] = entity
        elif isinstance(entity, VolumeEntity):
            assert not self.volumes.get(entity.nb), 'Volume nb '+str(entity.nb)+' already exists!'
            self.volumes[entity.nb] = entity

    def addEntities(self, entities):
        for entity in entities:
            self.addEntity(entity)


class Entity(object):
    """Base class for all entities.
    (!) Should not be created manually

    Parameters
    ----------
    nb: int
        Entity number.
    group: Optional[PhysicalGroup]
        Physical group of Entity.
    name: Optional[str]
        Name of Entity.
    mesh: Optional[py2gmsh.Mesh.Mesh]
        Mesh of entity.
    """
    __slots__ = ['nb', 'name' 'PhysicalGroup']
    def __init__(self, nb, group=None, name=None, mesh=None):
        self.nb = nb
        self.name = name
        self.PhysicalGroup = group
        if mesh is not None:
            mesh.addEntity(self)
        if group is not None:
            group.addEntity(self)

    def check_instance(self, entities, entity_class, index, mesh):
        if index is False and mesh is not None:
            for entity in entities:
                assert isinstance(entity, entity_class), 'points must be class instances of '+str(entity_class)
        elif index is True and mesh is not None:
        #    assert isinstance(mesh, Mesh), 'Mesh instance must be passed when defining entities per index'
            for entity in entities:
                assert isinstance(entity, int), 'index must be integers'



# POINTS

class Point(Entity):
    """Creates a Point.

    Parameters
    ----------
    xyz: array_like
        Coordinates of point (array of length 3)
    nb: Optional[int]
        Point number. If not set, it will use the class count 
        number (incrementing automatically upon creation of new
        points).
    group: Optional[PhysicalGroup]
        Physical group of Entity.
    mesh: Optional[py2gmsh.Mesh.Mesh]
        Mesh of entity.
    """
    def __init__(self, xyz, nb=None, group=None, mesh=None):
        super(Point, self).__init__(nb=nb, group=group, name='Point', mesh=mesh)
        self.xyz = xyz

    def setCoords(self, xyz):
        self.xyz[:] = [xyz]

    def _val2str(self):
        return '{'+str([v for v in self.xyz])[1:-1]+'}'


# CURVES

class CurveEntity(Entity):
    """Parent class for all curve type entities.

    Parameters
    ----------
    nb: Optional[int]
        Point number. If not set, it will use the class count 
        number (incrementing automatically upon creation of new
        curves).
    group: Optional[PhysicalGroup]
        Physical group of Entity.
    name: Optional[str]
        Name of curve.
    mesh: Optional[py2gmsh.Mesh.Mesh]
        Mesh of curve.
    """
    def __init__(self, nb=None, group=None, name=None, mesh=None):
        super(CurveEntity, self).__init__(nb=nb, group=group, name=name, mesh=mesh)

class Curve(CurveEntity):
    """Creates a Curve.

    Parameters
    ----------
    points: array_like[Points]
        Points instances (array of length 2)
    nb: Optional[int]
        CurveEntity number. If not set, it will use the class count 
        number (incrementing automatically upon creation of new
        points).
    group: Optional[PhysicalGroup]
        Physical group of entity.
    mesh: Optional[py2gmsh.Mesh.Mesh]
        Mesh of entity.
    """
    def __init__(self, points, nb=None, group=None, index=False, mesh=None):
        assert len(points) == 2, 'points array must be of length 2 when creating a curve'
        self.check_instance(points, Point, index, mesh)
        super(Curve, self).__init__(nb=nb, group=group, name='Curve', mesh=mesh)
        self.points = points
        self._index = index

    def setPoints(self, points):
        self.points[:] = [points]

    def _val2str(self):
        return '{'+str([v.nb for v in self.points])[1:-1]+'}'


Line = Curve


class Circle(CurveEntity):
    """Creates a Circle.

    Parameters
    ----------
    start: Point
        start point of circle arc.
    center: Point
        center point of circle arc.
    end: Point
        end point of circle arc.
    nb: Optional[int]
        CurveEntity number. If not set, it will use the class count 
        number (incrementing automatically upon creation of new
        points).
    group: Optional[PhysicalGroup]
        Physical group of entity.
    index: Optional[bool]
        Look for points defining the circle per index if True.
    mesh: Optional[py2gmsh.Mesh.Mesh]
        Mesh of entity.
    """
    def __init__(self, start, center, end, nb=None, group=None, index=False, mesh=None):
        self.check_instance([start, center, end], Point, index, mesh)
        super(Circle, self).__init__(nb=nb, group=group, name='Circle', mesh=mesh)
        self.start = start
        self.center = center
        self.end = end
        self._index = index

    def _val2str(self):
        return '{'+str([self.start.nb, self.center.nb, self.end.nb])[1:-1]+'}'


class CatmullRom(CurveEntity):
    def __init__(self, points, nb=None, group=None, index=False, mesh=None):
        self.check_instance(points, Point, index, mesh)
        super(Spline, self).__init__(nb=nb, group=group, name='CatmullRom', mesh=mesh)
        self.points = points
        self._index = index

    def _val2str(self):
        return '{'+str([v.nb for v in self.points])[1:-1]+'}'


class Ellipse(CurveEntity):
    def __init__(self, start, center, axis_point, end, nb=None, group=None, index=False, mesh=None):
        self.check_instance([start, center, acis_point, end], Point, index, mesh)
        super(Circle, self).__init__(nb=nb, group=group, name='Ellipse', mesh=mesh)
        self.start = start
        self.center = center
        self.end = end
        self.axis_point = axis_point
        self._index = index

    def _val2str(self):
        return '{'+str([self.start, self.center, self.axis_point, self.end])[1:-1]+'}'


class BSpline(CurveEntity):
    def __init__(self, points, nb=None, group=None, index=False, mesh=None):
        self.check_instance(points, Point, index, mesh)
        super(BSpline, self).__init__(nb=nb, group=group, name='BSpline', mesh=mesh)
        self.points = points
        self._index = index

    def _val2str(self):
        return '{'+str([v.nb for v in self.points])[1:-1]+'}'


class Spline(CurveEntity):
    def __init__(self, points, nb=None, group=None, index=False, mesh=None):
        self.check_instance(points, Point, index, mesh)
        super(Spline, self).__init__(nb=nb, group=group, name='Spline', mesh=mesh)
        self.points = points
        self._index = index

    def _val2str(self):
        return '{'+str([v.nb for v in self.points])[1:-1]+'}'


class CompoundCurve(CurveEntity):
    def __init__(self, curves, nb=None, group=None, index=False, mesh=None):
        self.check_instance(curves, Curve, index, mesh)
        super(Spline, self).__init__(nb=nb, group=group, name='Compound Curve', mesh=mesh)
        self.curves = curves
        self._index = index

    def _val2str(self):
        return '{'+str([v.nb for v in self.curves])[1:-1]+'}'


CompoundLine = CompoundCurve


# CURVE LOOPS

class CurveLoop(Entity):
    def __init__(self, curves, nb=None, group=None, index=False, mesh=None):
        self.check_instance(curves, Curve, index, mesh)
        super(CurveLoop, self).__init__(nb=nb, group=group, name='Curve Loop', mesh=mesh)
        self.curves = curves
        self._index = index

    def setCurves(self, curves):
        self.curves[:] = [curves]

    def _val2str(self):
        ll = []
        if self._index is False:
            for i, curve in enumerate(self.curves):
                if self.curves[i-1].points[1] == self.curves[i].points[0]:
                    ll += [self.curves[i].nb]
                elif self.curves[i-1].points[1] == self.curves[i].points[1]:
                    ll += [-self.curves[i].nb]
                elif self.curves[i-1].points[0] == self.curves[i].points[0]:
                    ll += [self.curves[i].nb]
                elif self.curves[i-1].points[0] == self.curves[i].points[1]:
                    ll += [-self.curves[i].nb]
                else:
                    assert 2<3, 'curveloop is wrong'
        else:
            ll = self.curves
        return '{'+str(ll)[1:-1]+'}'


LineLoop = CurveLoop


# SURFACES

class SurfaceEntity(Entity):
    """
    Parent class of all surface type entities
    """
    def __init__(self, nb=None, group=None, name=None, mesh=None):
        super(SurfaceEntity, self).__init__(nb=nb, group=group, name=name, mesh=mesh)


class PlaneSurface(SurfaceEntity):
    def __init__(self, curveloops, nb=None, group=None, index=False, mesh=None):
        self.check_instance(curveloops, CurveLoop, index, mesh)
        super(PlaneSurface, self).__init__(nb=nb, group=group, name='Plane Surface', mesh=mesh)
        self.curveloops = curveloops
        self._index = index

    def setCurveLoops(self, curveloops):
        self.curveloops[:] = [curveloops]

    def _val2str(self):
        return '{'+str([v.nb for v in self.curveloops])[1:-1]+'}'


class RuledSurface(SurfaceEntity):
    def __init__(self, curveloops, nb=None, group=None, sphere=None, mesh=None):
        self.check_instance(curveloops, CurveLoop, index, mesh)
        super(RuledSurface, self).__init__(nb=nb, group=group, name='Ruled Surface', mesh=mesh)
        self.curveloops = curveloops
        self.Sphere = sphere

    def setCurveLoops(self, curveloops):
        self.curveloops[:] = [curveloops]

    def _val2str(self):
        return '{'+str([v.nb for v in self.curveloops])[1:-1]+'}'


class CompoundSurface(SurfaceEntity):
    def __init__(self, surfaces, nb=None, group=None, index=False, mesh=None):
        self.check_instance(surfaces, SurfaceEntity, index, mesh)
        super(Spline, self).__init__(nb=nb, group=group, name='Compound Surface', mesh=mesh)
        self.surfaces = surfaces
        self._index = index

    def _val2str(self):
        return '{'+str([v.nb for v in self.surfaces])[1:-1]+'}'


# SURFACE LOOPS

class SurfaceLoop(Entity):
    def __init__(self, surfaces, nb=None, group=None, index=False, mesh=None):
        self.check_instance(surfaces, SurfaceLoop, index, mesh)
        super(SurfaceLoop, self).__init__(nb=nb, group=group, name='Surface Loop', mesh=mesh)
        self.surfaces = surfaces
        self._index = index

    def setSurfaces(self, surfaces):
        self.surfaces[:] = [surfaces]

    def _val2str(self):
        return '{'+str([v.nb for v in self.surfaces])[1:-1]+'}'


# VOLUMES

class VolumeEntity(Entity):
    """
    Parent class of all surface type entities
    """
    def __init__(self, nb=None, group=None, name=None, mesh=None):
        super(VolumeEntity, self).__init__(nb=nb, group=group, name=name, mesh=mesh)


class Volume(VolumeEntity):
    def __init__(self, surfaceloops, nb=None, group=None, index=False, mesh=None):
        self.check_instance(surfaceloops, SurfaceLoop, index, mesh)
        super(Volume, self).__init__(nb=nb, group=group, name='Volume', mesh=mesh)
        self.surfaceloops = surfaceloops
        self._index = index

    def _val2str(self):
        return '{'+str([v.nb for v in self.surfaceloops])[1:-1]+'}'

    def setSurfaceLoops(self, surfaceloops):
        self.surfaceloops[:] = [surfaceloops]


class CompoundVolume(VolumeEntity):
    def __init__(self, volumes, nb=None, group=None, index=False, mesh=None):
        self.check_instance(volumes, Volume, index, mesh)
        super(Volume, self).__init__(nb=nb, group=group, name='Compound Volume', mesh=mesh)
        self.volumes = volumes
        self._index = index

    def _val2str(self):
        return '{'+str([v.nb for v in self.volumes])[1:-1]+'}'
 



class OptionsHolder:
    def __init__(self):
        self.Mesh = Mesh()
        self.General = General()
        self.Geometry = Geometry()


class General:
    def __init__(self):
        self.AlphaBlending = None
        self.Antialiasing = None
        self.ArrowHeadRadius = None
        self.ArrowStemLength = None
        self.ArrowStemRadius = None
        self.Axes = None
        self.AxesAutoPosition = None
        self.AxesForceValue = None
        self.AxesFormatX = None
        self.AxesFormatY = None
        self.AxesFormatZ = None
        self.AxesLabelX = None
        self.AxesLabelY = None
        self.AxesLabelZ = None
        self.AxesMaxX = None
        self.AxesMaxY = None
        self.AxesMaxZ = None
        self.AxesMikado = None
        self.AxesMinX = None
        self.AxesMinY = None
        self.AxesMinZ = None
        self.AxesTicsX = None
        self.AxesTicsY = None
        self.AxesTicsZ = None
        self.AxesValueMaxX = None
        self.AxesValueMaxY = None
        self.AxesValueMaxZ = None
        self.AxesValueMinX = None
        self.AxesValueMinY = None
        self.AxesValueMinZ = None
        self.BackgroundGradient = None
        self.BackgroundImage3D = None
        self.BackgroundImageFileName = None
        self.BackgroundImageHeight = None
        self.BackgroundImagePage = None
        self.BackgroundImagePositionX = None
        self.BackgroundImagePositionY = None
        self.BackgroundImageWidth = None
        self.BoundingBoxSize = None
        self.Camera = None
        self.CameraAperture = None
        self.CameraEyeSeparationRatio = None
        self.CameraFocalLengthRatio = None
        self.Clip0A = None
        self.Clip0B = None
        self.Clip0C = None
        self.Clip0D = None
        self.Clip1A = None
        self.Clip1B = None
        self.Clip1C = None
        self.Clip1D = None
        self.Clip2A = None
        self.Clip2B = None
        self.Clip2C = None
        self.Clip2D = None
        self.Clip3A = None
        self.Clip3B = None
        self.Clip3C = None
        self.Clip3D = None
        self.Clip4A = None
        self.Clip4B = None
        self.Clip4C = None
        self.Clip4D = None
        self.Clip5A = None
        self.Clip5B = None
        self.Clip5C = None
        self.Clip5D = None
        self.ClipFactor = None
        self.ClipOnlyDrawIntersectingVolume = None
        self.ClipOnlyVolume = None
        self.ClipPositionX = None
        self.ClipPositionY = None
        self.ClipWholeElements = None
        self.Color = ColorGeneral()
        self.ColorScheme = None
        self.ConfirmOverwrite = None
        self.ContextPositionX = None
        self.ContextPositionY = None
        self.DefaultFileName = None
        self.DetachedMenu = None
        self.Display = None
        self.DisplayBorderFactor = None
        self.DoubleBuffer = None
        self.DrawBoundingBoxes = None
        self.ErrorFileName = None
        self.ExecutableFileName = None
        self.ExpertMode = None
        self.ExtraHeight = None
        self.ExtraPositionX = None
        self.ExtraPositionY = None
        self.ExtraWidth = None
        self.FastRedraw = None
        self.FieldHeight = None
        self.FieldPositionX = None
        self.FieldPositionY = None
        self.FieldWidth = None
        self.FileChooserPositionX = None
        self.FileChooserPositionY = None
        self.FileName = None
        self.FltkColorScheme = None
        self.FltkTheme = None
        self.FontSize = None
        self.GraphicsFont = None
        self.GraphicsFontEngine = None
        self.GraphicsFontSize = None
        self.GraphicsFontSizeTitle = None
        self.GraphicsFontTitle = None
        self.GraphicsHeight = None
        self.GraphicsPositionX = None
        self.GraphicsPositionY = None
        self.GraphicsWidth = None
        self.HighOrderToolsPositionX = None
        self.HighOrderToolsPositionY = None
        self.HighResolutionGraphics = None
        self.HighResolutionPointSizeFactor = None
        self.InitialModule = None
        self.Light0 = None
        self.Light0W = None
        self.Light0X = None
        self.Light0Y = None
        self.Light0Z = None
        self.Light1 = None
        self.Light1W = None
        self.Light1X = None
        self.Light1Y = None
        self.Light1Z = None
        self.Light2 = None
        self.Light2W = None
        self.Light2X = None
        self.Light2Y = None
        self.Light2Z = None
        self.Light3 = None
        self.Light3W = None
        self.Light3X = None
        self.Light3Y = None
        self.Light3Z = None
        self.Light4 = None
        self.Light4W = None
        self.Light4X = None
        self.Light4Y = None
        self.Light4Z = None
        self.Light5 = None
        self.Light5W = None
        self.Light5X = None
        self.Light5Y = None
        self.Light5Z = None
        self.LineWidth = None
        self.ManipulatorPositionX = None
        self.ManipulatorPositionY = None
        self.MaxX = None
        self.MaxY = None
        self.MaxZ = None
        self.MenuHeight = None
        self.MenuPositionX = None
        self.MenuPositionY = None
        self.MenuWidth = None
        self.MeshDiscrete = None
        self.MessageFontSize = None
        self.MessageHeight = None
        self.MinX = None
        self.MinY = None
        self.MinZ = None
        self.MouseHoverMeshes = None
        self.MouseSelection = None
        self.NonModalWindows = None
        self.NoPopup = None
        self.OptionsFileName = None
        self.OptionsPositionX = None
        self.OptionsPositionY = None
        self.Orthographic = None
        self.PluginHeight = None
        self.PluginPositionX = None
        self.PluginPositionY = None
        self.PluginWidth = None
        self.PointSize = None
        self.PolygonOffsetAlwaysOn = None
        self.PolygonOffsetFactor = None
        self.PolygonOffsetUnits = None
        self.ProgressMeterStep = None
        self.QuadricSubdivisions = None
        self.RecentFile0 = None
        self.RecentFile1 = None
        self.RecentFile2 = None
        self.RecentFile3 = None
        self.RecentFile4 = None
        self.RecentFile5 = None
        self.RecentFile6 = None
        self.RecentFile7 = None
        self.RecentFile8 = None
        self.RecentFile9 = None
        self.RotationCenterGravity = None
        self.RotationCenterX = None
        self.RotationCenterY = None
        self.RotationCenterZ = None
        self.RotationX = None
        self.RotationY = None
        self.RotationZ = None
        self.SaveOptions = None
        self.SaveSession = None
        self.ScaleX = None
        self.ScaleY = None
        self.ScaleZ = None
        self.SessionFileName = None
        self.Shininess = None
        self.ShininessExponent = None
        self.SmallAxes = None
        self.SmallAxesPositionX = None
        self.SmallAxesPositionY = None
        self.SmallAxesSize = None
        self.StatisticsPositionX = None
        self.StatisticsPositionY = None
        self.Stereo = None
        self.SystemMenuBar = None
        self.Terminal = None
        self.TextEditor = None
        self.TmpFileName = None
        self.Tooltips = None
        self.Trackball = None
        self.TrackballHyperbolicSheet = None
        self.TrackballQuaternion0 = None
        self.TrackballQuaternion1 = None
        self.TrackballQuaternion2 = None
        self.TrackballQuaternion3 = None
        self.TranslationX = None
        self.TranslationY = None
        self.TranslationZ = None
        self.VectorType = None
        self.Verbosity = None
        self.VisibilityPositionX = None
        self.VisibilityPositionY = None
        self.WatchFilePattern = None
        self.ZoomFac = None

class Geometry:
    def __init__(self):
        self.AutoCoherence = None
        self.Clip = None
        self.Color = ColorGeometry()
        self.CopyDisplayAttributes = None
        self.CopyMeshingMethod = None
        self.DoubleClickedEntityTag = None
        self.DoubleClickedLineCommand = None
        self.DoubleClickedPointCommand = None
        self.DoubleClickedSurfaceCommand = None
        self.DoubleClickedVolumeCommand = None
        self.ExactExtrusion = None
        self.ExtrudeReturnLateralEntities = None
        self.ExtrudeSplinePoints = None
        self.HideCompounds = None
        self.HighlightOrphans = None
        self.LabelType = None
        self.Light = None
        self.LightTwoSide = None
        self.LineNumbers = None
        self.Lines = None
        self.LineSelectWidth = None
        self.LineType = None
        self.LineWidth = None
        self.MatchGeomAndMesh = None
        self.Normals = None
        self.NumSubEdges = None
        self.OCCConnectFaces = None
        self.OCCFixDegenerated = None
        self.OCCFixSmallEdges = None
        self.OCCFixSmallFaces = None
        self.OCCScaling = None
        self.OCCSewFaces = None
        self.OffsetX = None
        self.OffsetY = None
        self.OffsetZ = None
        self.OldCircle = None
        self.OldNewReg = None
        self.OldRuledSurface = None
        self.OrientedPhysicals = None
        self.PointNumbers = None
        self.Points = None
        self.PointSelectSize = None
        self.PointSize = None
        self.PointType = None
        self.ScalingFactor = None
        self.SnapX = None
        self.SnapY = None
        self.SnapZ = None
        self.SurfaceNumbers = None
        self.Surfaces = None
        self.SurfaceType = None
        self.Tangents = None
        self.Tolerance = None
        self.Transform = None
        self.TransformXX = None
        self.TransformXY = None
        self.TransformXZ = None
        self.TransformYX = None
        self.TransformYY = None
        self.TransformYZ = None
        self.TransformZX = None
        self.TransformZY = None
        self.TransformZZ = None
        self.VolumeNumbers = None
        self.Volumes = None

class Mesh:
    def __init__(self):
        self.Algorithm = None
        self.Algorithm3D = None
        self.AllowSwapAngle = None
        self.AngleSmoothNormals = None
        self.AnisoMax = None
        self.BdfFieldFormat = None
        self.Binary = None
        self.CgnsImportOrder = None
        self.ChacoArchitecture = None
        self.ChacoEigensolver = None
        self.ChacoEigTol = None
        self.ChacoGlobalMethod = None
        self.ChacoHypercubeDim = None
        self.ChacoLocalMethod = None
        self.ChacoMeshDim1 = None
        self.ChacoMeshDim2 = None
        self.ChacoMeshDim3 = None
        self.ChacoParamINTERNAL_VERTICES = None
        self.ChacoParamREFINE_MAP = None
        self.ChacoParamREFINE_PARTITION = None
        self.ChacoParamTERMINAL_PROPOGATION = None
        self.ChacoPartitionSection = None
        self.ChacoSeed = None
        self.ChacoVMax = None
        self.CharacteristicLengthExtendFromBoundary = None
        self.CharacteristicLengthFactor = None
        self.CharacteristicLengthFromCurvature = None
        self.CharacteristicLengthFromPoints = None
        self.CharacteristicLengthMax = None
        self.CharacteristicLengthMin = None
        self.Clip = None
        self.Color = ColorMesh()
        self.ColorCarousel = None
        self.CpuTime = None
        self.DoRecombinationTest = None
        self.DrawSkinOnly = None
        self.Dual = None
        self.ElementOrder = None
        self.Explode = None
        self.FlexibleTransfinite = None
        self.Format = None
        self.Hexahedra = None
        self.HighOrderNumLayers = None
        self.HighOrderOptimize = None
        self.HighOrderOptPrimSurfMesh = None
        self.HighOrderPoissonRatio = None
        self.HighOrderThresholdMax = None
        self.HighOrderThresholdMin = None
        self.IgnorePartitionBoundary = None
        self.LabelSampling = None
        self.LabelType = None
        self.LcIntegrationPrecision = None
        self.Light = None
        self.LightLines = None
        self.LightTwoSide = None
        self.LineNumbers = None
        self.Lines = None
        self.LineWidth = None
        self.Lloyd = None
        self.MeshOnlyVisible = None
        self.MetisAlgorithm = None
        self.MetisEdgeMatching = None
        self.MetisRefinementAlgorithm = None
        self.MinimumCirclePoints = None
        self.MinimumCurvePoints = None
        self.MshFilePartitioned = None
        self.MshFileVersion = None
        self.NbHexahedra = None
        self.NbNodes = None
        self.NbPartitions = None
        self.NbPrisms = None
        self.NbPyramids = None
        self.NbQuadrangles = None
        self.NbTetrahedra = None
        self.NbTriangles = None
        self.NbTrihedra = None
        self.NewtonConvergenceTestXYZ = None
        self.Normals = None
        self.NumSubEdges = None
        self.OldRefinement = None
        self.Optimize = None
        self.OptimizeNetgen = None
        self.Partitioner = None
        self.PartitionHexWeight = None
        self.PartitionPrismWeight = None
        self.PartitionPyramidWeight = None
        self.PartitionQuadWeight = None
        self.PartitionTetWeight = None
        self.PartitionTrihedronWeight = None
        self.PartitionTriWeight = None
        self.PointNumbers = None
        self.Points = None
        self.PointSize = None
        self.PointType = None
        self.PreserveNumberingMsh2 = None
        self.Prisms = None
        self.Pyramids = None
        self.Quadrangles = None
        self.QualityInf = None
        self.QualitySup = None
        self.QualityType = None
        self.RadiusInf = None
        self.RadiusSup = None
        self.RandomFactor = None
        self.RecombinationAlgorithm = None
        self.RecombinationTestHorizStart = None
        self.RecombinationTestNoGreedyStrat = None
        self.Recombine3DAll = None
        self.Recombine3DConformity = None
        self.Recombine3DLevel = None
        self.RecombineAll = None
        self.RefineSteps = None
        self.RemeshAlgorithm = None
        self.RemeshParametrization = None
        self.SaveAll = None
        self.SaveElementTagType = None
        self.SaveGroupsOfNodes = None
        self.SaveParametric = None
        self.ScalingFactor = None
        self.SecondOrderExperimental = None
        self.SecondOrderIncomplete = None
        self.SecondOrderLinear = None
        self.SmoothCrossField = None
        self.Smoothing = None
        self.SmoothNormals = None
        self.SmoothRatio = None
        self.SubdivisionAlgorithm = None
        self.SurfaceEdges = None
        self.SurfaceFaces = None
        self.SurfaceNumbers = None
        self.SwitchElementTags = None
        self.Tangents = None
        self.Tetrahedra = None
        self.ToleranceEdgeLength = None
        self.ToleranceInitialDelaunay = None
        self.Triangles = None
        self.Trihedra = None
        self.VolumeEdges = None
        self.VolumeFaces = None
        self.VolumeNumbers = None
        self.Voronoi = None
        self.ZoneDefinition = None


class ColorMesh:
    def __init__(self):
        self.Eight = None
        self.Eighteen = None
        self.Eleven = None
        self.Fifteen = None
        self.Five = None
        self.Four = None
        self.Fourteen = None
        self.Hexahedra = None
        self.Lines = None
        self.Nine = None
        self.Nineteen = None
        self.Normals = None
        self.One = None
        self.Points = None
        self.PointsSup = None
        self.Prisms = None
        self.Pyramids = None
        self.Quadrangles = None
        self.Seven = None
        self.Seventeen = None
        self.Six = None
        self.Sixteen = None
        self.Tangents = None
        self.Ten = None
        self.Tetrahedra = None
        self.Thirteen = None
        self.Three = None
        self.Triangles = None
        self.Trihedra = None
        self.Twelve = None
        self.Two = None
        self.Zero = None


class ColorGeneral:
    def __init__(self):
        self.AmbientLight = None
        self.Axes = None
        self.Background = None
        self.BackgroundGradient = None
        self.DiffuseLight = None
        self.Foreground = None
        self.SmallAxes = None
        self.SpecularLight = None
        self.Text = None


class ColorGeometry:
    def __init__(self):
        self.HighlightOne = None
        self.HighlightTwo = None
        self.HighlightZero = None
        self.Lines = None
        self.Normals = None
        self.Points = None
        self.Projection = None
        self.Selection = None
        self.Surfaces = None
        self.Tangents = None
        self.Volumes = None
